/*
 * Copyright (C) 2005-2012 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file write.c Paraslash's standalone wav/raw player. */

#include <regex.h>
#include <sys/types.h>

#include "para.h"
#include "string.h"
#include "write.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "stdin.h"
#include "buffer_tree.h"
#include "write.h"
#include "write_common.h"
#include "fd.h"
#include "error.h"
#include "version.h"
#include "check_wav.h"

INIT_WRITE_ERRLISTS;

static struct write_args_info conf;

static struct stdin_task sit;

static int loglevel;
INIT_STDERR_LOGGING(loglevel)

__noreturn static void print_help_and_die(void)
{
	int d = conf.detailed_help_given;
	const char **p = d? write_args_info_detailed_help
		: write_args_info_help;

	printf_or_die("%s\n\n", WRITE_CMDLINE_PARSER_PACKAGE "-"
		WRITE_CMDLINE_PARSER_VERSION);
	printf_or_die("%s\n\n", write_args_info_usage);
	for (; *p; p++)
		printf_or_die("%s\n", *p);
	print_writer_helps(d);
	exit(0);
}

/*
 * Parse config and register a task for a writer node.
 *
 * \param arg Command line arguments.
 * \param parent The new node will be a child of \a parent.
 * \param wn The writer node.
 *
 * If arg is \p NULL, the OS-dependent default writer is used with no
 * arguments.  The default writers are alsa for Linux, osx for OS X, oss for
 * *BSD, and the file writer if the default writer is not supported.
 *
 * Once the writer configuration has been retrieved from the ->parse_config
 * callback a writer node is created, its buffer tree node is added to the
 * buffer tree as a child of the given parent.
 *
 * Finally, the new writer node's task structure is initialized and registered
 * to the paraslash scheduler.
 *
 * \return Standard.
 */
static void setup_writer_node(const char *arg, struct btr_node *parent,
		struct writer_node *wn, struct sched *s)
{
	wn->conf = check_writer_arg_or_die(arg, &wn->writer_num);
	register_writer_node(wn, parent, s);
}

struct write_task {
	struct task task;
	struct check_wav_context *cwc;
};

static void write_pre_select(struct sched *s, struct task *t)
{
	struct write_task *wt = container_of(t, struct write_task, task);
	check_wav_pre_select(s, wt->cwc);
}

static void write_post_select(__a_unused struct sched *s, struct task *t)
{
	struct write_task *wt = container_of(t, struct write_task, task);
	t->error = check_wav_post_select(wt->cwc);
}

static int setup_and_schedule(void)
{
	int i, ret;
	struct btr_node *cw_btrn;
	struct writer_node *wns;
	static struct sched s;
	struct wav_params wp;
	struct write_task wt = {
		.task = {
			.pre_select = write_pre_select,
			.post_select = write_post_select,
			.status = "write task",
		},
	};

	loglevel = get_loglevel_by_name(conf.loglevel_arg);
	sit.btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "stdin"));
	stdin_set_defaults(&sit);
	register_task(&s, &sit.task);

	COPY_WAV_PARMS(&wp, &conf);
	wt.cwc = check_wav_init(sit.btrn, NULL, &wp, &cw_btrn);
	register_task(&s, &wt.task);
	if (!conf.writer_given) {
		wns = para_calloc(sizeof(*wns));
		setup_writer_node(NULL, cw_btrn, wns, &s);
		i = 1;
	} else {
		wns = para_calloc(conf.writer_given * sizeof(*wns));
		for (i = 0; i < conf.writer_given; i++)
			setup_writer_node(conf.writer_arg[i], cw_btrn,
				wns + i, &s);
	}

	s.default_timeout.tv_sec = 10;
	s.default_timeout.tv_usec = 50000;
	ret = schedule(&s);
	if (ret >= 0) {
		int j;
		for (j = 0; j < i; j++) {
			struct task *t = &wns[j].task;
			assert(t->error < 0);
			if (t->error != -E_WRITE_COMMON_EOF
					&& t->error != -E_BTR_EOF) {
				PARA_ERROR_LOG("%s: %s\n", t->status,
					para_strerror(-t->error));
				if (ret >= 0)
					ret = t->error;
			}
		}
	}
	for (i--; i >= 0; i--) {
		struct writer_node *wn = wns + i;
		struct writer *w = writers + wn->writer_num;

		w->close(wn);
		btr_remove_node(&wn->btrn);
		w->free_config(wn->conf);
		free(wn->conf);
	}
	free(wns);
	check_wav_shutdown(wt.cwc);
	return ret;
}

/**
 * Para_write's main function.
 *
 * \param argc The usual argument counter.
 * \param argv The usual argument vector.
 *
 * It sets up and starts the tasks and the buffer tree nodes determined by
 * command line options.
 *
 * \return \p EXIT_SUCCESS or EXIT_FAILURE
 */
int main(int argc, char *argv[])
{
	int ret;

	writer_init();
	write_cmdline_parser(argc, argv, &conf);
	HANDLE_VERSION_FLAG("write", conf);
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();

	ret = setup_and_schedule();
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		exit(EXIT_FAILURE);
	}
	exit(EXIT_SUCCESS);
}
